# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module for creating the visualizations for better understanding the results obtained
through the benchmark module.
"""
import os

import numpy as np
from matplotlib import pyplot as plt

from ..types import Dict, Tuple, Optional


class BenchmarkDisplay:
    """
    A class for generating the visualizations required to properly interpret the benchmarks'
    results.
    """

    @staticmethod
    def load_bench_result(path: str) -> Dict[str, Tuple[np.array, np.array, float]]:
        """
        Loads an evaluation's or whole benchmark's results.

        Parameters
        ----------
        path
            A string with the path to the desired results.

        Returns
        -------
        results
            A dictionary with the experiments' names and their corresponding results (curves, mean curve, roc)
        """
        result = np.load(os.path.join(path), allow_pickle=True)
        return result

    @staticmethod
    def plot_bench_from_path(path: str, path_to_save: str = None) -> None:
        """
        Loads the results from a file on disk, plots it and optionally saves the figure to the disk.

        Parameters
        ----------
        path
            A string with the file from which to load the results.
        path_to_save
            An (optional) string with the path onto which to save the figure.
        """
        result = BenchmarkDisplay.load_bench_result(path)
        BenchmarkDisplay.plot_bench(result, path_to_save)

    @staticmethod
    def plot_bench(
            result: Dict[str, Tuple[np.array, np.array, float]],
            path_to_save: Optional[str] = None,
            title: Optional[str] = None
    ) -> None:
        """
        Plots the results from a whole benchmark on a single figure, and optionally saves the
        figure to the disk.

        Parameters
        ----------
        result
            A dictionary with the experiments' names and their results.
        path_to_save
            An (optional) string with the path onto which to save the generated figure.
            If None, it only shows the figure, otherwise, it only saves the figure but doesn't show it.
        title
            An (optional) string specifying the figure's title.
        """
        fig, axs = plt.subplots(nrows=1, ncols=len(result), figsize=(20, 10))

        if len(result) == 1:
            axs = [axs]

        if title is not None:
            fig.suptitle(title, y=0.99)
        fig.subplots_adjust(top=0.8)

        for i, (name, (curves, mean_curve, roc)) in enumerate(result.items()):
            curve_length = len(mean_curve)
            valid_curvs = []
            for curve in curves:
                if not np.isnan(curve[0]):
                    valid_curvs.append(curve)
                    axs[i].plot(np.linspace(0., 1., curve_length), curve, 'C0', alpha=0.25)
            mean_curv = np.mean(np.asarray(valid_curvs), axis=0)
            axs[i].plot(np.linspace(0., 1., curve_length), mean_curv, 'C0')
            roc = np.mean(mean_curve)

            axs[i].plot(np.linspace(0., 1., curve_length), np.linspace(0., 1., curve_length), 'C1')
            axs[i].set_title(
                f'Mislabeled detection {name} \n ROC={roc} \n Nbr of run={len(valid_curvs)}')
            axs[i].set_xlabel('Part of the dataset searched')
            axs[i].set_ylabel('Part of mislabeled found')
            axs[i].grid('minor')
        if path_to_save is None:
            plt.show()
        else:
            plt.savefig(path_to_save)



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing a Nearest Neighbors interface and a Linear Nearest Neighbors
algorithm. It will prove itself useful for finding the top-k most influential
examples of datasets, as implemented in the influence calculator interface.
"""
from abc import abstractmethod

import tensorflow as tf

from .sorted_dict import BatchSort, ORDER
from ..types import Callable, Optional, Tuple


class BaseNearestNeighbors:
    """
    A Nearest Neighbors interface for efficiently searching for specific data-points
    in a dataset.
    """

    @abstractmethod
    def build(
        self,
        dataset: tf.data.Dataset,
        dot_product_fun: Callable[[tf.Tensor, tf.Tensor], tf.Tensor],
        k: int,
        query_batch_size: int,
        d_type: tf.DType = tf.float32,
        order: ORDER = ORDER.DESCENDING
    ) -> None:
        """
        Builds the nearest neighbors object that will be used to find the k neighbors
        inside the provided dataset.

        Parameters
        ----------
        dataset
            A TF dataset containing the points which shall be indexed.
        dot_product_fun
            The dot product function used to compute the distance between 2 points
        k
            An integer for the amount of samples to search for
        query_batch_size
            An integer for the query's batch size
        d_type
            The dataset's element's data-type
        order
            Either descending or ascending for the top or bottom results as per the similarity metric
        """
        raise NotImplementedError()

    @abstractmethod
    def query(self, vector_to_find: tf.Tensor, batch_size: Optional[int] = None) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Find the k closest points to the provided vector in the object's dataset.

        Parameters
        ----------
        vector_to_find
            A tensor of points to query.
        batch_size
            An integer with the query's batch size

        Returns
        -------
        values_and_samples
            A tuple with the k closest points to the queries and their corresponding distances in the format
            (distances, points)
        """
        raise NotImplementedError()


class LinearNearestNeighbors(BaseNearestNeighbors):
    """
    An implementation of a Linear Nearest Neighbors search algorithm using a given similarity metric and
    doing as much lazy computations as possible for scalability.

    Attributes
    ----------
    dataset
        A TF dataset with the points from which the nearest neighbors will be searched.
    dot_product_fun
        A callable taking in two vectors and returning a notion of similarity between them.
    batched_sorted_dict
        A BatchSort instance that takes care of keeping the top/bottom most similar examples from
        the batch.
    """

    def __init__(self):
        self.dataset = None
        self.dot_product_fun = None
        self.batched_sorted_dict = None

    def build(
        self,
        dataset: tf.data.Dataset,
        dot_product_fun: Callable[[tf.Tensor, tf.Tensor], tf.Tensor],
        k: int,
        query_batch_size: int,
        d_type: tf.DType = tf.float32,
        order: ORDER = ORDER.DESCENDING
        ) -> None:
        """
        Builds the linear nearest neighbors object that will be used to find the k neighbors
        inside the provided dataset.

        Parameters
        ----------
        dataset
            A TF dataset containing the points which shall be indexed.
        dot_product_fun
            The dot product function used to compute the distance between 2 points
        k
            An integer for the amount of samples to search for
        query_batch_size
            An integer for the query's batch size
        d_type
            The dataset's element's data-type
        order
            Either descending or ascending for the top or bottom results as per the similarity metric
        """
        self.dataset = dataset
        self.dot_product_fun = dot_product_fun
        batch_shape = self.dataset.element_spec[0][0].shape[1:]
        self.batched_sorted_dict = BatchSort(batch_shape, [query_batch_size, k], dtype=d_type, order=order)

    def query(self, vector_to_find: tf.Tensor, batch_size: Optional[int] = None) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Find the k closest points to the provided vector in the object's dataset.

        Parameters
        ----------
        vector_to_find
            A tensor of points to query.
        batch_size
            An integer with the query's batch size

        Returns
        -------
        values_and_samples
            A tuple with the k closest points to the queries and their corresponding distances in the format
            (distances, points)
        """
        if batch_size is None:
            batch_size = tf.shape(vector_to_find)[0]

        dataset_iterator = iter(self.dataset)
        self.batched_sorted_dict.reset()

        def body_func(i):
            batch, ihvp = next(dataset_iterator)
            influence_values = self.dot_product_fun(vector_to_find, ihvp)

            self.batched_sorted_dict.add_all(
                tf.repeat(tf.expand_dims(batch[0], axis=0), batch_size, axis=0),
                influence_values
            )

            return (i+1, )

        tf.while_loop(
            cond=lambda i: i < self.dataset.cardinality(),
            body=body_func,
            loop_vars=[tf.constant(0, dtype=tf.int64)]
        )

        training_samples, influences_values = self.batched_sorted_dict.get()

        return influences_values, training_samples



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Backtracking Line Search optimizer algorithm with Stochastic Gradient Descent steps.
This code was based on an implementation by Louis Bethune (ANITI) -- https://github.com/Algue-Rythme
"""
from dataclasses import dataclass

import tensorflow as tf
from tensorflow.keras import Model # pylint: disable=E0611
from tensorflow.keras.optimizers import Optimizer, SGD # pylint: disable=E0611

from ..types import Callable


@dataclass()
class BTLSParameters:
    """
    Data class containing the Backtracking Line Search optimizer's internal parameters
    """
    beta: float
    gamma: float
    eta: float
    max_eta: float
    min_eta: float


class BacktrackingLineSearch(Optimizer):
    """
    Implementation of a batched Backtracking Line Search optimizer with SGD steps.

    Parameters
    ----------
    batches_per_epoch
        An integer indicating the amount of batches that constitute a whole epoch. This information is used for scaling
        the optimizer updates.
    scaling_factor
        A scaling factor for the Wolfe condition
    """
    def __init__(
            self,
            batches_per_epoch: int,
            scaling_factor: float,
            **kwargs
    ):
        super().__init__(name="backtracking_line_search", **kwargs)
        self.optimizer = SGD()
        self.scaling_factor = scaling_factor
        self.batches_per_epoch = int(batches_per_epoch)
        self.parameters = BTLSParameters(
            beta=0.9,
            gamma=2 ** (1 / self.batches_per_epoch),
            eta=1.,
            max_eta=10.,
            min_eta=1e-6
        )

    def step(self, model, current_loss, x_inputs, labels, gradients):
        """
        Performs a step of the line-search optimizer by attempting stochastic gradient descents until the Wolfe
        condition is met.

        Parameters
        ----------
        model
            A compiled TF model with accessible weights
        current_loss
            A tensor with the current loss value for the batch
        x_inputs
            The inputs for the batch
        labels
            The corresponding labels for the batch
        gradients
            The gradients of the loss wrt the model's weights for the provided loss tensor
        """
        # Save the original weights for the batch
        curr_weights = model.get_weights()

        # Attempt a first direction
        norm = self.c_gradnorm(gradients)
        def closure():
            return model.compiled_loss(labels, model(x_inputs, training=True))
        # closure = lambda: model.compiled_loss(labels, model(x_inputs, training=True))
        self.parameters.eta *= self.parameters.gamma
        direction = self.attempt_step(model, curr_weights, gradients, closure)

        # Repeat progressively smaller steps until the (approximate) Wolfe condition is verified
        while not BacktrackingLineSearch.wolfe_condition(direction, current_loss, norm, self.parameters.eta):
            self.parameters.eta *= self.parameters.beta
            if self.parameters.max_eta < self.parameters.eta < self.parameters.min_eta:
                break
            direction = self.attempt_step(model, curr_weights, gradients, closure)

        # Save the most suitable learning rate in the permitted range
        self.parameters.eta = min(max(self.parameters.min_eta, self.parameters.eta), self.parameters.max_eta)

    @staticmethod
    def wolfe_condition(target, source, norm, eta):
        """
        Verifies whether the Wolfe condition is being met for a set of loss values (before vs after an SGD step).

        Parameters
        ----------
        target
            A tensor with the value of the loss at the current inner step of the line-search
        source
            A tensor with the value of the loss at the beginning of the optimizer's step
        norm
            A tensor with the norm of the gradients for the current loss
        eta
            The current step size in the inner step of the line-search

        Returns
        -------
        wolfe_condition
            A boolean indicating whether the Wolfe condition is verified for the current learning rate
        """
        return tf.less_equal(target, source - eta * norm)

    def attempt_step(self, model: Model, curr_weights: tf.Tensor, gradients: tf.Tensor, closure: Callable):
        """
        Performs a step of SGD using the updated learning rate in the direction of the gradient and returns the new
        value of the loss function with the new weights.

        Parameters
        ----------
        model
            A TF model with accessible weights and variables
        curr_weights
            A tensor with the current model's weights
        gradients
            A tensor with the loss' gradients wrt the model's weights on the batch
        closure
            An object whose call returns the loss function's value using the updated weights

        Returns
        -------
        called_closure
            The new value of the loss for the model with the updated weights
        """
        self.optimizer.learning_rate.assign(self.parameters.eta)
        model.set_weights(curr_weights)
        self.optimizer.apply_gradients(zip(gradients, model.trainable_variables))
        return closure()

    def c_gradnorm(self, gradients):
        """
        Compute the squared norm of the loss function's gradients

        Parameters
        ----------
        gradients
            A tensor with the loss function's gradients

        Returns
        -------
        norm
            A tensor with the norm of the loss function's gradients scaled by the parameter scaling_factor
        """
        return self.scaling_factor * tf.linalg.global_norm(gradients) ** 2

    def _resource_apply_dense(self, grad, handle, apply_state):
        pass

    def _resource_apply_sparse(self, grad, handle, indices, apply_state):
        pass

    def get_config(self):
        """
        Implementation of configuration setter method, as required for the Optimizer interface.
        """
        base_config = super().get_config()
        base_config["batches_per_epoch"] = self.batches_per_epoch
        base_config["scaling_factor"] = self.scaling_factor
        return base_config



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Custom wrappers for tensorflow model
"""
import itertools

import tensorflow as tf
from tensorflow.keras.losses import Reduction  # pylint: disable=E0611

from ..utils import assert_batched_dataset, from_layer_name_to_layer_idx, default_process_batch
from ..types import Callable, Optional, Union, List, Tuple

ProcessBatchTypeAlias = Callable[[Tuple[tf.Tensor, ...]], Tuple[tf.Tensor, tf.Tensor, tf.Tensor]]


class BaseInfluenceModel:
    """
    A generic Tensorflow model wrapper for Influence functions that facilitates the access to the
    weights for which these quantities are to be computed.

    Attributes
    ----------
    model
        Model used for computing influence score.
    weights_to_watch
        List of the model weights to watch when computing gradients, jacobians & hessians.
    loss_function
        Loss function to calculate influence (e.g. keras CategoricalCrossentropy). Make sure not to
        apply any reduction (Reduction.NONE), and specify correctly if the output is `from_logits`
        for example.
    process_batch_for_loss_fn
        A callable for preprocessing the batch to transform it into a format that can be treated
        by the algorithm: (inputs, label, sample weight).
    weights_processed
        A boolean indicating whether the weights are already in the right format to be ingested by
        the class' methods or if they need to be processed beforehand.
    """

    def __init__(self,
                 model: tf.keras.Model,
                 weights_to_watch: Optional[List[tf.Variable]] = None,
                 loss_function: Callable = tf.keras.losses.CategoricalCrossentropy(
                    from_logits=False, reduction=Reduction.NONE),
                 process_batch_for_loss_fn: ProcessBatchTypeAlias = default_process_batch,
                 weights_processed: bool = False):

        if hasattr(loss_function, 'reduction') and loss_function.reduction is not Reduction.NONE:
            raise ValueError('The loss function must not have reduction.')

        self.model = model
        self.weights_processed = weights_processed
        if weights_to_watch is None:
            weights_to_watch = [layer.weights for layer in model.layers]
            self.weights_processed = False
        # "flatten" the list of weights and remove empty weights
        self.weights = self.__process_weights_list(weights_to_watch)

        self.nb_params = tf.reduce_sum([tf.size(w) for w in self.weights])
        self.loss_function = loss_function
        self.process_batch_for_loss_fn = process_batch_for_loss_fn

    def __call__(self, inps: tf.Tensor) -> tf.Tensor:
        """
        Computes the forward of the original model

        Parameters
        ----------
        inps
            Inputs on which to make the inference.

        Returns
        -------
        y
            Outputs of the original model.
        """
        return self.model(inps)

    def __process_weights_list(self, weights_to_watch: Union[tf.Tensor, List[tf.Tensor]]) -> List[tf.Tensor]:
        """
        Ensure a proper formatting of the weights
        TODO: Improve it, cause list(itertools.chain(*weights_to_watch)) is not idempotent

        Parameters
        ----------
        weights_to_watch
            A collection of weights for which we wish to compute the influence function in some unspecified
            format

        Returns
        -------
        processed_weights
            A list with the weights in the right format for them to be ingested by the rest of the class' methods
        """
        if self.weights_processed:
            return weights_to_watch

        return list(itertools.chain(*weights_to_watch))

    @property
    def layers(self):
        """
        Access the layers of the original model.

        Returns
        -------
        layers
            The layers of the original model.
        """
        return self.model.layers

    @staticmethod
    @tf.function
    def _loss(
        model: tf.keras.Model,
        loss_function: Callable,
        batch: Tuple[tf.Tensor, ...],
        process_batch_for_loss_fn: ProcessBatchTypeAlias
    ) -> tf.Tensor:
        """
        Computes the model's loss for a single batch of samples.

        Parameters
        ----------
        model
            The model used for computing the influence score.
        loss_function
            The reduction-less loss function to calculate the influence (e.g. cross-entropy).
        batch
            A batch of tuples of sets of inputs and their corresponding outputs.
        process_batch_for_loss_fn
            A callable for preprocessing the batch to transform it into a format that can be treated
            by the algorithm: (inputs, label, sample weight).

        Returns
        -------
        loss_values
            The loss values for each input (i.e. not reduced).
        """
        model_inp, y_true, sample_weight = process_batch_for_loss_fn(batch)
        return loss_function(y_true, model(model_inp), sample_weight)

    @staticmethod
    @tf.function
    def _jacobian(model: tf.keras.Model, weights: tf.Tensor, loss_function: Callable,
                  batch: Tuple[tf.Tensor, ...], process_batch_for_loss_fn: ProcessBatchTypeAlias) -> tf.Tensor:
        """
        Computes the model's jacobian for a single batch of samples.

        Parameters
        ----------
        model
            The model used for computing the influence score.
        loss_function
            The reduction-less loss function to calculate the influence (e.g. cross-entropy).
        batch
            A batch of tuples of sets of inputs and their corresponding outputs.
        process_batch_for_loss_fn
            A callable for preprocessing the batch to transform it into a format that can be treated
            by the algorithm: (inputs, label, sample weight).

        Returns
        -------
        jacobian
            The jacobian matrix for the set of inputs.
        """
        model_inp, y_true, sample_weight = process_batch_for_loss_fn(batch)
        batch_size = tf.shape(y_true)[0]
        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(weights)
            y_pred = loss_function(y_true, model(model_inp), sample_weight)

        jacobian = tape.jacobian(y_pred, weights)

        jacobian = [tf.reshape(j, (batch_size, -1,)) for j in jacobian]
        jacobian = tf.concat(jacobian, axis=1)

        return jacobian

    @staticmethod
    @tf.function
    def _gradient(model: tf.keras.Model, weights: tf.Variable, loss_function: Callable,
                  batch: Tuple[tf.Tensor, ...], process_batch_for_loss_fn: ProcessBatchTypeAlias) -> tf.Tensor:
        """
        Computes the model gradients for a single batch of sample.

        Parameters
        ----------
        model
            The model used for computing the influence score.
        loss_function
            Reduction-less loss function to calculate the influence (e.g. cross-entropy).
        batch
            A batch of tuples of sets of inputs and their corresponding outputs.
        process_batch_for_loss_fn
            A callable for preprocessing the batch to transform it into a format that can be treated
            by the algorithm: (inputs, label, sample weight).

        Returns
        -------
        gradient
            The gradient vector for the set of inputs.
        """
        model_inp, y_true, sample_weight = process_batch_for_loss_fn(batch)
        with tf.GradientTape(watch_accessed_variables=False) as tape:
            tape.watch(weights)
            y_pred = tf.expand_dims(loss_function(y_true, model(model_inp), sample_weight), axis=-1)

        gradients = tape.gradient(y_pred, weights)
        # note that it is the accumulated gradients for all inputs in the batch
        gradients = [tf.reshape(g, (-1,)) for g in gradients]
        gradients = tf.concat(gradients, axis=0)

        return gradients

    def _loss_tensor(self, batch: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the model's loss on the batched tensor

        Parameters
        ----------
        batch
            A batch of tuples of sets of inputs and their corresponding outputs.

        Returns
        -------
        loss_values
            Loss values for each of the points of the batch.
        """
        loss_values = BaseInfluenceModel._loss(self.model, self.loss_function, batch, self.process_batch_for_loss_fn)

        return loss_values

    def batch_loss(self, dataset: tf.data.Dataset) -> tf.Tensor:
        """
        Computes the model's loss on the whole batched dataset.

        Parameters
        ----------
        dataset
            Batched dataset on which to compute the loss.

        Returns
        -------
        loss_values
            Loss values for each of the points in the dataset.
        """
        assert_batched_dataset(dataset)

        loss_values = tf.concat([
            BaseInfluenceModel._loss(self.model, self.loss_function, batch, self.process_batch_for_loss_fn)
            for batch in dataset
        ], axis=0)

        return loss_values

    @tf.function
    def batch_jacobian_tensor(self, batch: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the jacobian of the loss wrt the weights of the start_layer on a Tensor

        Parameters
        ----------
        batch
            A batch of tuples of sets of inputs and their corresponding outputs.

        Returns
        -------
        jacobians
            Matrix of the first-order partial derivative of the loss function wrt the
            start_layer weights.
        """
        jacobians = BaseInfluenceModel._jacobian(self.model, self.weights, self.loss_function,
                                             batch, self.process_batch_for_loss_fn)

        return jacobians

    def batch_jacobian(self, dataset: tf.data.Dataset) -> tf.Tensor:
        """
        Computes the jacobian of the loss wrt the weights of the start_layer on the whole
        batched dataset.

        Parameters
        ----------
        dataset
            Batched dataset on which to compute the jacobian.

        Returns
        -------
        jacobians
            Matrix of the first-order partial derivative of the loss function wrt the
            start_layer weights.
        """
        assert_batched_dataset(dataset)

        jacobians = tf.concat([
            BaseInfluenceModel._jacobian(self.model, self.weights, self.loss_function,
                                         batch, self.process_batch_for_loss_fn)
            for batch in dataset
        ], axis=0)

        return jacobians

    @tf.function
    def batch_gradient_tensor(self, batch: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the gradient of the loss wrt the weights of the start_layer on a Tensor

        Parameters
        ----------
        batch
            A batch of tuples of sets of inputs and their corresponding outputs.

        Returns
        -------
        gradients
            Gradient values of the loss function wrt the start_layer's weights.
        """
        gradients = BaseInfluenceModel._gradient(self.model, self.weights, self.loss_function,
                                             batch, self.process_batch_for_loss_fn)

        return gradients

    def batch_gradient(self, dataset) -> tf.Tensor:
        """
        Computes the gradient of the loss wrt the weights of the start_layer on the whole
        batched dataset.

        Parameters
        ----------
        dataset
            Batched dataset on which to compute the gradient.

        Returns
        -------
        gradients
            Gradient values of the loss function wrt the start_layer's weights.
        """
        assert_batched_dataset(dataset)

        gradients = tf.stack([
            BaseInfluenceModel._gradient(self.model, self.weights, self.loss_function,
                                         batch, self.process_batch_for_loss_fn)
            for batch in dataset
        ])

        return gradients


class InfluenceModel(BaseInfluenceModel):
    """
    A Tensorflow model wrapper for Influence functions which only require the first layer
    index or name from which we will watch the weights (e.g. one decides to ignore the feature extractor).

    Parameters
    ----------
    model
        Model used for computing influence score.
    start_layer
        Starting layer name or index for the weights and bias collection. If set to None,
        will search for the last layer with weights before logits.
    last_layer
        Last layer name or index for the weights and biases collection.
        If set to None, only the layer indicated in the start_layer parameter will be used.
    loss_function
        Loss function to calculate influence (e.g. keras CategoricalCrossentropy). Make sure not to
        apply any reduction (Reduction.NONE), and specify correctly if the output is `from_logits`
        for example.
    process_batch_for_loss_fn
        A callable for preprocessing the batch to transform it into a format that can be treated
        by the algorithm: (inputs, label, sample weight).
    """
    def __init__(self,
                 model: tf.keras.Model,
                 start_layer: Optional[Union[str, int]] = None,
                 last_layer: Optional[Union[str, int]] = None,
                 loss_function: Callable = tf.keras.losses.CategoricalCrossentropy(
                     from_logits=False, reduction=Reduction.NONE),
                 process_batch_for_loss_fn: ProcessBatchTypeAlias = default_process_batch):
        self.start_layer = start_layer
        weights_to_watch = InfluenceModel._get_weights_of_interest(model, start_layer, last_layer)
        super().__init__(model, weights_to_watch, loss_function, process_batch_for_loss_fn, weights_processed=True)

    @staticmethod
    def _get_weights_of_interest(model: tf.keras.Model,
                                 start_layer: Optional[Union[str, int]],
                                 last_layer: Optional[Union[str, int]]) -> list:
        """
        Gets the list of trainable weights from layer 'start_layer' to layer 'last_layer' in model

        Parameters
        ----------
        model
            Model we want to get the weights from.
        start_layer
            Starting layer for the weights and bias collection. If set to None, will search for the
            last layer with weights before logits.
        last_layer
            Last layer for the weights and biases collection used in hessian computation. If set to
            None, only the layer indicated in the start_layer parameter will be used.

        Returns
        -------
        weights
            A flatten list of weights between the start_layer and the last_layer layers in model.
        """
        # get an id value for the start_layer parameter
        if start_layer is None:
            start_layer = InfluenceModel._find_last_weight_layer(model)
            start_layer = len(model.layers) + start_layer
        elif isinstance(start_layer, str):
            start_layer = from_layer_name_to_layer_idx(model, start_layer)
        else:
            assert(isinstance(start_layer, int)), "start_layer should be None, a string or an int"

        # get the list of layers of interest
        if last_layer is None:
            layers_for_influence = [model.layers[start_layer]]
        elif isinstance(last_layer, str):
            last_layer = from_layer_name_to_layer_idx(model, last_layer)
            assert last_layer >= start_layer, \
                f"last_layer id: {last_layer} should be greater than start_layer id: {start_layer}"
            layers_for_influence = model.layers[start_layer : last_layer+1]
            start_layer = last_layer
        else:
            assert(isinstance(last_layer, int)), "last_layer should be None, a string or an int"
            if last_layer < 0:
                last_layer += len(model.layers)
                assert last_layer >= start_layer, \
                    f"last_layer id: {last_layer} should be greater than start_layer id: {start_layer}"
            elif last_layer == 0:
                assert last_layer == start_layer, \
                    f"last_layer id: {last_layer} should be greater than start_layer id: {start_layer}"
            layers_for_influence = model.layers[start_layer : last_layer+1]

        # get the list of weights of interest
        weights = [lay.weights for lay in layers_for_influence]
        weights = list(itertools.chain(*weights))
        return weights

    @staticmethod
    def _find_last_weight_layer(model: tf.keras.Model) -> int:
        """
        Find and return the id of the last layer before logits with weights.

        Parameters
        ----------
        model
            Model used for computing influence score.

        Returns
        -------
        layer_id
            Id (e.g. -2, -3...) of the layer found.
        """
        for layer_id in range(2, len(model.layers)):
            layer = model.layers[-layer_id]
            if hasattr(layer, 'weights') and layer.weights:
                return -layer_id
        raise ValueError('No layers with weights found for the model.')



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing the different approaches for computing and approximating the
inverse-hessian-vector product: an essential block in the computation of influence
functions.
"""
from abc import ABC, abstractmethod
from enum import Enum
from argparse import ArgumentError

import tensorflow as tf
from tensorflow.keras import Model  # pylint: disable=E0611
from tensorflow.keras.models import Sequential  # pylint: disable=E0611

from .model_wrappers import BaseInfluenceModel, InfluenceModel

from ..types import Optional, Union, Tuple, List, Callable
from ..utils import assert_batched_dataset, conjugate_gradients_solve, map_to_device


class InverseHessianVectorProduct(ABC):
    """
    An interface for classes that perform hessian-vector products.

    Parameters
    ----------
    model
       A TF model following the InfluenceModel interface whose weights we wish to use for the calculation of
       these (inverse)-hessian-vector products.
    train_dataset
       A batched TF dataset containing the training dataset's point we wish to employ for the estimation of
       the hessian matrix.
    """
    def __init__(self, model: InfluenceModel, train_dataset: Optional[tf.data.Dataset]):
        if train_dataset is not None:
            self.cardinality = train_dataset.cardinality()

        self.model = model
        self.train_set = train_dataset


    @abstractmethod
    def _compute_ihvp_single_batch(self, group_batch: Tuple[tf.Tensor, ...], use_gradient: bool = True) -> tf.Tensor:
        """
        Computes the inverse-hessian-vector product of a group of points provided in the form of
        a batch of tensors.

        Parameters
        ----------
        group_batch
            A Tuple with a single batch of tensors containing the points of which we wish to
            compute the inverse-hessian-vector product
        use_gradient
            A boolean indicating whether the IHVP is with the gradients wrt to the loss of the
            points in group or with these vectors instead.

        Returns
        -------
        ihvp
            A tensor containing a rank-1 tensor per input point.
        """
        raise NotImplementedError

    def compute_ihvp(self, group: tf.data.Dataset, use_gradient: bool = True) -> tf.data.Dataset:
        """
        Computes the inverse-hessian-vector product of a group of points.

        Parameters
        ----------
        group
            A TF dataset containing the group of points of which we wish to compute the
            inverse-hessian-vector product.
        use_gradient
            A boolean indicating whether the IHVP is with the gradients wrt to the loss of the
            points in group or with these vectors instead.

        Returns
        -------
        ihvp
            A tensor containing one rank-1 tensor per input point
        """
        assert_batched_dataset(group)

        ihvp_dataset = group.map(lambda *single_batch: self._compute_ihvp_single_batch(single_batch, use_gradient))

        return ihvp_dataset

    @abstractmethod
    def _compute_hvp_single_batch(self, group_batch: Tuple[tf.Tensor, ...], use_gradient: bool = True) -> tf.Tensor:
        """
        Computes the hessian-vector product of a group of points.

        Parameters
        ----------
        group_batch
            A TF dataset containing the group of points of which we wish to compute the
            hessian-vector product.
        use_gradient
            A boolean indicating whether the hvp is with the gradients wrt to the loss of the
            points in group or with these vectors instead.

        Returns
        -------
        hvp
            A tensor containing one rank-1 tensor per input point
        """
        raise NotImplementedError()

    def compute_hvp(self, group: tf.data.Dataset, use_gradient: bool = True) -> tf.data.Dataset:
        """
        Computes the hessian-vector product of a group of points.

        Parameters
        ----------
        group
            A TF dataset containing the group of points of which we wish to compute the
            hessian-vector product.
        use_gradient
            A boolean indicating whether the hvp is with the gradients wrt to the loss of the
            points in group or with these vectors instead.

        Returns
        -------
        hvp
            A tensor containing one rank-1 tensor per input point
        """
        assert_batched_dataset(group)

        hvp_ds = group.map(lambda *single_batch: self._compute_hvp_single_batch(single_batch, use_gradient))

        return hvp_ds


class ExactIHVP(InverseHessianVectorProduct):
    """
    A class that performs the 'exact' computation of the inverse-hessian-vector product.
    As such, it will calculate the hessian of the provided model's loss wrt its parameters,
    compute its Moore-Penrose pseudo-inverse (for numerical stability) and the multiply it
    by the gradients.

    Notes
    -----
    To speed up the algorithm, the hessian matrix is calculated once at instantiation.

    For models with a considerable amount of weights, this implementation may be infeasible
    due to its O(n^2) complexity for the hessian, plus the O(n^3) for its inversion.
    If its memory consumption is too high, you should consider using the CGD approximation,
    or computing the hessian separately and initializing the ExactIHVP with this hessian while
    setting train_dataset to None. To expect it to work the hessian should be computed for
    the training_set.

    Parameters
    ----------
    model
        The TF2.X model implementing the InfluenceModel interface.
    train_dataset
        The TF dataset, already batched and containing only the samples we wish to use for
        the computation of the hessian matrix. Either train_hessian or train_dataset should
        not be None but not both.
    train_hessian
        The estimated hessian matrix of the model's loss wrt its parameters computed with
        the samples used for the model's training. Either hessian or train_dataset should
        not be None but not both.
    """
    def __init__(
            self,
            model: InfluenceModel,
            train_dataset: Optional[tf.data.Dataset] = None,
            train_hessian: Optional[tf.Tensor] = None,
    ):
        super().__init__(model, train_dataset)
        if train_dataset is not None:
            nb_batch = tf.cast(train_dataset.cardinality(), dtype=tf.int32)
            self.inv_hessian = self._compute_inv_hessian(self.train_set, nb_batch)
            self.hessian = None
        elif train_hessian is not None:
            self.hessian = train_hessian
            self.inv_hessian = tf.linalg.pinv(train_hessian)
        else:
            raise ArgumentError("Either train_dataset or train_hessian can be set to None, but not both")

    @tf.function
    def _compute_inv_hessian(self, dataset: tf.data.Dataset, nb_batch: tf.int32) -> tf.Tensor:
        """
        Compute the (pseudo)-inverse of the hessian matrix wrt to the model's parameters using
        backward-mode AD.

        Disclaimer
        ----------
        This implementation trades memory usage for speed, so it can be quite
        memory intensive, especially when dealing with big models.

        Parameters
        ----------
        dataset
            A TF dataset containing the whole or part of the training dataset for the
            computation of the inverse of the mean hessian matrix.

        Returns
        ----------
        inv_hessian
            A tf.Tensor with the resulting inverse hessian matrix
        """
        weights = self.model.weights

        hess = tf.zeros((self.model.nb_params, self.model.nb_params), dtype=dataset.element_spec[0].dtype)
        nb_elt = tf.constant(0, dtype=tf.int32)
        nb_batch_saw = tf.constant(0, dtype=tf.int32)
        iter_ds = iter(dataset)

        def hessian_sum(nb_elt, nb_batch_saw, hess):
            batch = next(iter_ds)
            nb_batch_saw += tf.constant(1, dtype=tf.int32)
            curr_nb_elt = tf.shape(batch[0])[0]
            nb_elt += curr_nb_elt
            with tf.GradientTape(persistent=True, watch_accessed_variables=False) as tape_hess:
                tape_hess.watch(weights)
                grads = self.model.batch_jacobian_tensor(batch) # pylint: disable=W0212

            curr_hess = tape_hess.jacobian(
                    grads, weights
                    )

            curr_hess = [tf.reshape(h, shape=(len(grads), self.model.nb_params, -1)) for h in curr_hess]
            curr_hess = tf.concat(curr_hess, axis=-1)
            curr_hess = tf.reshape(curr_hess, shape=(len(grads), self.model.nb_params, -1))
            curr_hess = tf.reduce_sum(curr_hess, axis=0)
            hess += tf.cast(curr_hess, dtype=hess.dtype)

            return nb_elt, nb_batch_saw, hess

        nb_elt, _, hess = tf.while_loop(
            cond=lambda __, nb_batch_saw, _: nb_batch_saw < nb_batch,
            body=hessian_sum,
            loop_vars=[nb_elt, nb_batch_saw, hess]
        )

        hessian = hess / tf.cast(nb_elt, dtype=hess.dtype)

        return tf.linalg.pinv(hessian)

    @tf.function
    def _compute_ihvp_single_batch(self, group_batch: Tuple[tf.Tensor, ...], use_gradient: bool = True) -> tf.Tensor:
        """
        Computes the inverse-hessian-vector product of a group of points provided in the form of
        a batch of tensors by computing the exact inverse hessian matrix and performing the product
        operation.

        Parameters
        ----------
        group_batch
            A Tuple with a single batch of tensors containing the points of which we wish to
            compute the inverse-hessian-vector product.
        use_gradient
            A boolean indicating whether the IHVP is with the gradients wrt to the loss of the
            points in group or with these vectors instead.

        Returns
        -------
        ihvp
            A tensor containing a rank-1 tensor per input point.
        """
        if use_gradient:
            grads = tf.reshape(self.model.batch_jacobian_tensor(group_batch), (-1, self.model.nb_params))
        else:
            grads = tf.reshape(group_batch[0], (-1, self.model.nb_params))

        ihvp = tf.matmul(self.inv_hessian, tf.cast(grads, dtype=self.inv_hessian.dtype), transpose_b=True)
        return ihvp

    @tf.function
    def _compute_hvp_single_batch(self, group_batch: Tuple[tf.Tensor, ...], use_gradient: bool = True) -> tf.Tensor:
        """
        Computes the hessian-vector product of a group of points provided in the form of a tuple
        of tensors by computing the hessian matrix and performing the product operation.

        Parameters
        ----------
        group_batch
            A Tuple with a single batch of tensors containing the points of which we wish to
            compute the hessian-vector product.
        use_gradient
            A boolean indicating whether the hvp is with the gradients wrt to the loss of the
            points in group or with these vectors instead.

        Returns
        -------
        hvp
            A tensor containing one rank-1 tensor per input point
        """
        if use_gradient:
            grads = tf.reshape(self.model.batch_jacobian_tensor(group_batch), (-1, self.model.nb_params))
        else:
            grads = tf.reshape(group_batch[0], (-1, self.model.nb_params))

        hvp = tf.matmul(self.hessian, grads, transpose_b=True)

        return hvp

    def compute_hvp(self, group: tf.data.Dataset, use_gradient: bool = True) -> tf.data.Dataset:
        """
        Computes the hessian-vector product of a group of points provided in the form of a tuple
        of tensors by computing the hessian matrix and performing the product operation.

        Parameters
        ----------
        group
            A Tuple with a single batch of tensors containing the points of which we wish to
            compute the hessian-vector product.
        use_gradient
            A boolean indicating whether the hvp is with the gradients wrt to the loss of the
            points in group or with these vectors instead.

        Returns
        -------
        hvp
            A tensor containing one rank-1 tensor per input point
        """
        if self.hessian is None:
            self.hessian = tf.linalg.pinv(self.inv_hessian)
        return super().compute_hvp(group, use_gradient)


class ForwardOverBackwardHVP:
    """
    A class for efficiently computing Hessian-vector products using forward-over-backward
    auto-differentiation.
    This module is used for the approximate IHVP calculations (CGD and LISSA).

    Parameters
    ----------
    model
        A TF model following the InfluenceModel interface.
    train_dataset
        A (batched) TF dataset with the data-points that will be used for the hessian.
    weights
        The target weights on which to calculate the HVP.
    """
    def __init__(
            self,
            model: BaseInfluenceModel,
            train_dataset: tf.data.Dataset,
            weights: Optional[List[tf.Tensor]] = None
    ):
        self.model = model
        self.train_dataset = train_dataset
        self.cardinality = train_dataset.cardinality()

        if weights is None:
            self.weights = model.weights
        else:
            self.weights = weights

    @staticmethod
    def _reshape_vector(grads: tf.Tensor, weights: tf.Tensor) -> List[tf.Tensor]:
        """
        Reshapes the gradient vector to the right shape for being input into the HVP computation.

        Parameters
        ----------
        grads
            A tensor with the computed gradients.
        weights
            A tensor with the target weights.

        Returns
        -------
        grads_reshape
            A list with the gradients in the right shape.
        """
        grads_reshape = []
        index = 0
        for w in weights:
            shape = tf.shape(w)
            size = tf.reduce_prod(shape)
            g = grads[index:(index + size)]
            grads_reshape.append(tf.reshape(g, shape))
            index += size
        return grads_reshape

    @tf.function
    def _sub_call(
            self,
            x: tf.Tensor,
            feature_maps_hessian_current: tf.Tensor,
            y_hessian_current: tf.Tensor
    ) -> tf.Tensor:
        """
        Performs the hessian-vector product for a single feature map.

        Parameters
        ----------
        x
            The gradient vector to be multiplied by the hessian matrix.
        feature_maps_hessian_current
            The current feature map for the hessian calculation.
        y_hessian_current
            The label corresponding to the current feature map.

        Returns
        -------
        hessian_vector_product
            A tf.Tensor containing the result of the hessian-vector product for a given input point and one pair
            feature map-label.
        """
        with tf.autodiff.ForwardAccumulator(
                self.weights,
                # The "vector" in Hessian-vector product.
                x) as acc:
            with tf.GradientTape(persistent=False, watch_accessed_variables=False) as tape:
                tape.watch(self.weights)
                loss = self.model.loss_function(y_hessian_current, self.model(feature_maps_hessian_current))
            backward = tape.jacobian(loss, self.weights)
        hessian_vector_product = acc.jvp(backward)

        hvp = [tf.reshape(hessian_vp, shape=(-1,)) for hessian_vp in hessian_vector_product]
        hvp = tf.concat(hvp, axis=0)

        weight = tf.cast(tf.shape(feature_maps_hessian_current)[0], dtype=hvp.dtype)

        hvp = hvp * weight

        return hvp

    def __call__(self, x_initial: tf.Tensor) -> tf.Tensor:
        """
        Computes the mean hessian-vector product for a given feature map over a set of points.

        Parameters
        ----------
        x_initial
            The point of the dataset over which this product will be computed

        Returns
        -------
        hessian_vector_product
            Tensor with the hessian-vector product
        """
        x = self._reshape_vector(x_initial, self.model.weights)

        hvp_init = tf.zeros((self.model.nb_params,), dtype=x_initial.dtype)
        dataset_iterator = iter(self.train_dataset)

        def body_func(i, hessian_vector_product, nb_hessian):
            features_block, labels_block = next(dataset_iterator)

            def batched_hvp(elt):
                f, l = elt
                hessian_product_current = self._sub_call(x, tf.expand_dims(f, axis=0), tf.expand_dims(l, axis=0))

                return hessian_product_current

            hessian_vector_product_inner = tf.reduce_sum(
                tf.map_fn(fn=batched_hvp, elems=[features_block, labels_block], fn_output_signature=x_initial.dtype),
                axis=0
            )

            hessian_vector_product += hessian_vector_product_inner
            return i + 1, hessian_vector_product, nb_hessian + tf.shape(features_block)[0]

        _, hessian_vector_product, nb_hessian = tf.while_loop(
            cond=lambda i, _, __: i < self.cardinality,
            body=body_func,
            loop_vars=[tf.constant(0, dtype=tf.int64), hvp_init, tf.constant(0, dtype=tf.int32)]
        )

        hessian_vector_product = tf.reshape(hessian_vector_product, (self.model.nb_params, 1)) / \
                                 tf.cast(nb_hessian, dtype=hessian_vector_product.dtype)

        return hessian_vector_product


class IterativeIHVP(InverseHessianVectorProduct):
    """
    A class that approximately computes inverse-hessian-vector products leveraging forward-over-backward
    automatic differentiation with an iterative procedure to estimate the product directly, without needing to
    calculate the hessian matrix or invert it.
    Notes
    -----
    It is ideal for models containing a considerable amount of parameters. It does however trade memory for
    speed, as the calculations for estimating the inverse hessian operator are repeated for each sample.
    Parameters
    ----------
    iterative_function
        The procedure to compute the inverse hessian product operation
    model
        The TF2.X model implementing the InfluenceModel interface
    extractor_layer
        An integer indicating the position of the last layer of the feature extraction network.
    train_dataset
        The TF dataset, already batched and containing only the samples we wish to use for the computation of the
        hessian matrix
    n_opt_iters
        The maximum amount of CGD iterations to perform when estimating the inverse-hessian
    feature_extractor
        If the feature extraction model is not Sequential, the full TF graph must be provided for the computation of
        the different feature maps.
    """
    def __init__(
            self,
            iterative_function,
            model: InfluenceModel,
            extractor_layer: Union[int, str],
            train_dataset: tf.data.Dataset,
            n_opt_iters: Optional[int] = 100,
            feature_extractor: Optional[Model] = None,
    ):
        super().__init__(model, train_dataset)
        self.n_opt_iters = n_opt_iters
        self._batch_shape_tensor = None
        self.extractor_layer = extractor_layer

        if feature_extractor is None:
            assert isinstance(model.model, Sequential)
            self.feature_extractor = tf.keras.Sequential(self.model.layers[:self.extractor_layer])
        else:
            assert isinstance(feature_extractor, Model)
            self.feature_extractor = feature_extractor

        self.train_set = self._compute_feature_map_dataset(self.train_set)  # extract the train set's features
        self.model = BaseInfluenceModel(
            tf.keras.Sequential(model.layers[extractor_layer:]),
            weights_to_watch=model.weights,
            loss_function=model.loss_function,
            weights_processed=True
        )  # model that predicts based on the extracted feature maps
        self.weights = self.model.weights
        self.hessian_vector_product = ForwardOverBackwardHVP(self.model, self.train_set, self.weights)
        self.iterative_function = iterative_function

    def batch_shape_tensor(self):
        """
        Return the batch shape of a tensor
        """
        return self._batch_shape_tensor

    def _compute_feature_map_dataset(self, dataset: tf.data.Dataset) -> tf.data.Dataset:
        """
        Extracts the feature maps for an entire dataset and creates a TF dataset associating them with
        their corresponding labels.
        Parameters
        ----------
        dataset
            The TF dataset whose feature maps we wish to extract using the model's first layers
        Returns
        -------
        feature_map_dataset
            A TF dataset with the pairs (feature_maps, labels), batched using the same batch_size as the one provided
            as input
        """
        feature_map_dataset = map_to_device(dataset, lambda x_batch, y: (self.feature_extractor(x_batch), y)).cache()

        if self._batch_shape_tensor is None:
            self._batch_shape_tensor = tf.shape(next(iter(feature_map_dataset))[0])

        return feature_map_dataset

    def _compute_ihvp_single_batch(self, group_batch: Tuple[tf.Tensor, ...], use_gradient: bool = True) -> tf.Tensor:
        """
        Computes the inverse-hessian-vector product of a group of points provided in the form of
        a batch of tensors by inverting the hessian-vector product that is calculated through
        forward-over-backward AD.
        Parameters
        ----------
        group_batch
            A Tuple with a single batch of tensors containing the points of which we wish to
            compute the inverse-hessian-vector product.
        use_gradient
            A boolean indicating whether the IHVP is with the gradients wrt to the loss of the
            points in group or with these vectors instead.
        Returns
        -------
        ihvp
            A tensor containing a rank-1 tensor per input point.
        """
        # Transform the dataset into a set of feature maps-labels
        if use_gradient:
            feature_maps = self.feature_extractor(group_batch[0])
            grads = self.model.batch_jacobian_tensor((feature_maps, *group_batch[1:]))
        else:
            grads = tf.reshape(group_batch[0], (-1, self.model.nb_params))

        # Compute the IHVP for each pair feature map-label
        def cgd_func(single_grad):
            inv_hessian_vect_product = self.iterative_function(self.hessian_vector_product,
                                                               tf.expand_dims(single_grad, axis=-1),
                                                               self.n_opt_iters)
            return inv_hessian_vect_product

        ihvp_list = tf.map_fn(fn=cgd_func, elems=grads)

        ihvp_list = tf.transpose(ihvp_list) if ihvp_list.shape[-1] != 1 \
            else tf.transpose(tf.squeeze(ihvp_list, axis=-1))

        return ihvp_list

    def _compute_hvp_single_batch(self, group_batch: Tuple[tf.Tensor, ...], use_gradient: bool = True) -> tf.Tensor:
        """
        Computes the hessian-vector product of a group of points provided in the form of a tuple
        of tensors through forward-over-backward AD.
        Parameters
        ----------
        group_batch
            A Tuple with a single batch of tensors containing the points of which we wish to
            compute the hessian-vector product.
        use_gradient
            A boolean indicating whether the hvp is with the gradients wrt to the loss of the
            points in group or with these vectors instead.
        Returns
        -------
        hvp
            A tensor containing one rank-1 tensor per input point
        """
        # Transform the dataset into a set of feature maps-labels
        if use_gradient:
            feature_maps = self.feature_extractor(group_batch[0])
            grads = self.model.batch_jacobian_tensor((feature_maps, *group_batch[1:]))
        else:
            grads = tf.reshape(group_batch[0], (-1, self.model.nb_params))

        # Compute the HVP for each pair features map - label
        def single_hvp(single_grad):
            hvp = self.hessian_vector_product(tf.expand_dims(single_grad, axis=-1))
            return hvp

        hvp_list = tf.map_fn(fn=single_hvp, elems=grads)

        hvp_list = tf.transpose(hvp_list) if hvp_list.shape[-1] != 1 else tf.transpose(tf.squeeze(hvp_list, axis=-1))

        return hvp_list


class ConjugateGradientDescentIHVP(IterativeIHVP):
    """
    A class that approximately computes inverse-hessian-vector products leveraging forward-over-backward
    automatic differentiation and Conjugate Gradient Descent to estimate the product directly, without needing to
    calculate the hessian matrix or invert it.
    Notes
    -----
    It is ideal for models containing a considerable amount of parameters. It does however trade memory for
    speed, as the calculations for estimating the inverse hessian operator are repeated for each sample.
    Parameters
    ----------
    model
        The TF2.X model implementing the InfluenceModel interface
    extractor_layer
        An integer indicating the position of the last layer of the feature extraction network.
    train_dataset
        The TF dataset, already batched and containing only the samples we wish to use for the computation of the
        hessian matrix
    n_opt_iters
        The maximum amount of CGD iterations to perform when estimating the inverse-hessian
    feature_extractor
        If the feature extraction model is not Sequential, the full TF graph must be provided for the computation of
        the different feature maps.
    """
    def __init__(
            self,
            model: InfluenceModel,
            extractor_layer: Union[int, str],
            train_dataset: tf.data.Dataset,
            n_opt_iters: Optional[int] = 100,
            feature_extractor: Optional[Model] = None,
    ):
        def iterative_function(operator, v, maxiter):  # pylint: disable=W0613
            return conjugate_gradients_solve(operator, v, x0=None, maxiter=self.n_opt_iters)
        super().__init__(iterative_function, model, extractor_layer, train_dataset, n_opt_iters, feature_extractor)


class LissaIHVP(IterativeIHVP):
    """
    A class that approximately computes inverse-hessian-vector products leveraging forward-over-backward
    automatic differentiation and lissa [https://arxiv.org/pdf/1703.04730.pdf , https://arxiv.org/pdf/1602.03943.pdf]
    to estimate the product directly, without needing to calculate the hessian matrix or invert it.

    [A^{-1}v]_{j+1} = v + (I - (A + d * I))[A^{-1}v]_j * v

    Notes
    -----
    It is ideal for models containing a considerable amount of parameters. It does however trade memory for
    speed, as the calculations for estimating the inverse hessian operator are repeated for each sample.

    Parameters
    ----------
    model
        The TF2.X model implementing the InfluenceModel interface
    extractor_layer
        An integer indicating the position of the last layer of the feature extraction network.
    train_dataset
        The TF dataset, already batched and containing only the samples we wish to use for the computation of the
        hessian matrix
    n_opt_iters
        The maximum amount of CGD iterations to perform when estimating the inverse-hessian
    feature_extractor
        If the feature extraction model is not Sequential, the full TF graph must be provided for the computation of
        the different feature maps.
    damping
        A damping parameter to regularize a nearly singular operator.
    scale
        A rescaling factor to verify the hypothesis of norm(operator / scale) < 1.
    """
    def __init__(
            self,
            model: InfluenceModel,
            extractor_layer: Union[int, str],
            train_dataset: tf.data.Dataset,
            n_opt_iters: Optional[int] = 100,
            feature_extractor: Optional[Model] = None,
            damping: float = 1e-4,
            scale: float = 10.
    ):
        super().__init__(self.lissa, model, extractor_layer, train_dataset, n_opt_iters, feature_extractor)
        self.damping = tf.convert_to_tensor(damping, dtype=tf.float32)
        self.scale = tf.convert_to_tensor(scale, dtype=tf.float32)

    def lissa(self, operator: Callable, v: tf.Tensor, maxiter: int):
        """
        Performs the Linear time Stochastic Second-order Algorithm (LiSSA) optimization procedure to solve
        a problem of the shape Ax = b by iterating as follows:

            [A^{-1}v]_{j+1} = v + (I - (A + d * I))[A^{-1}v]_j * v

        Parameters
        ----------
        operator
            The operator that transforms the input vector v into Av
        v
            The vector v of the problem
        maxiter
            Number of iterations of the algorithm

        Returns
        -------
        ihvp_result
            A tensor containing inv(A)v
        """
        _, ihvp_result = tf.while_loop(lambda index, ihvp: index < maxiter,
                                       lambda index, ihvp: (index + 1,
                                                            v + tf.cast(1. - self.damping, dtype=tf.float32) * ihvp -
                                                            operator(ihvp) / self.scale),
                                       [tf.constant(0, dtype=tf.int32), v])
        ihvp_result /= self.scale

        return ihvp_result


class IHVPCalculator(Enum):
    """
    Inverse Hessian Vector Product Calculator interface.
    """
    Exact = ExactIHVP
    Cgd = ConjugateGradientDescentIHVP
    Lissa = LissaIHVP

    @staticmethod
    def from_string(ihvp_calculator: str) -> 'IHVPCalculator':
        """
        Restore an IHVPCalculator from string.

        Parameters
        ----------
        ihvp_calculator
            String indicated the method use to compute the inverse hessian vector product,
            e.g 'exact' or 'cgd'.

        Returns
        -------
        ivhp_calculator
            IHVPCalculator object.
        """
        assert ihvp_calculator in ['exact', 'cgd', 'lissa'], "Only 'exact', 'lissa' and 'cgd' inverse hessian " \
                                                             "vector product calculators are supported."
        if ihvp_calculator == 'exact':
            return IHVPCalculator.Exact
        if ihvp_calculator == 'lissa':
            return IHVPCalculator.Lissa

        return IHVPCalculator.Cgd



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
A module implementing a mislabeled sample detection on the CIFAR-10 image classifiction
dataset.
For each evaluation, a model will be trained on the noisy dataset, the self-influence
will be computed for each point, and they will get sorted by this value. A well
performing influence calculator technique should be able to put forth the mislabeled
samples among the top-most influential points.
"""
import ssl

import numpy as np
import tensorflow as tf
from tensorflow.keras import Model # pylint: disable=E0611
from tensorflow.keras.models import Sequential # pylint: disable=E0611
from tensorflow.keras.applications import EfficientNetB0, VGG19 # pylint: disable=E0611
from tensorflow.keras.layers import Dense, Flatten, Dropout # pylint: disable=E0611
from tensorflow.keras.regularizers import L1L2 # pylint: disable=E0611
from tensorflow.keras.losses import CategoricalCrossentropy # pylint: disable=E0611
from tensorflow.keras.optimizers import Adam # pylint: disable=E0611

from .base_benchmark import BaseTrainingProcedure, MislabelingDetectorEvaluator, ModelsSaver
from .model_resnet import ResNet

from ..types import Tuple, Union, Any, Optional, List

ssl._create_default_https_context = ssl._create_unverified_context # pylint: disable=W0212


class ConvNetCIFAR(Sequential):
    """
    A class with some convolutional neural network implementations on which to benchmark the
    different influence calculation techniques. The feature extractor will either be a
    ResNet18, an EfficientNetB0 or a VGG-19.

    Parameters
    ----------
    model
        Either a string for the base model ('resnet', 'efficient_net' or 'vgg19'), or a TF
        base model itself.
    use_regularization
        A boolean indicating whether to add regularization on the final model's last layer.
    """
    def __init__(self, model: Union[str, Model], use_regularization: bool = True, use_bias: bool = True, **kwargs):
        super().__init__(**kwargs)
        if isinstance(model, Model):
            base_model = model
        else:
            if model == 'resnet':
                base_model = ResNet(input_shape=(32, 32, 3), include_top=False, block='basic_block',
                                    residual_unit='v1',
                                    repetitions=(3, 3, 3), initial_filters=16, initial_pooling=None,
                                    final_pooling=None)
            elif model == 'efficient_net':
                base_model = EfficientNetB0(include_top=False, weights=None, input_shape=(32, 32, 3))
            elif model == 'vgg19':
                base_model = VGG19(include_top=False, weights=None, input_shape=(32, 32, 3))
            else:
                raise ValueError('unknown model=' + model)
        self.add(base_model)
        self.add(Flatten())

        if use_regularization:
            dense_1 = Dense(128, kernel_regularizer=L1L2(l1=1e-4, l2=1e-4))
        else:
            dense_1 = Dense(128)
        self.add(dense_1)

        self.add(Dropout(0.4))
        self.add(tf.keras.layers.LeakyReLU())

        if use_regularization:
            dense_2 = Dense(
                10,
                kernel_regularizer=L1L2(l1=1e-4, l2=1e-4),
                kernel_initializer="he_normal",
                use_bias=use_bias
            )
        else:
            dense_2 = Dense(10, use_bias=use_bias)

        self.add(dense_2)


class Cifar10TrainingProcedure(BaseTrainingProcedure):
    """
    A class for streamlining the process of training models on the CIFAR-10 dataset.

    Parameters
    ----------
    epochs
        An integer indicating for how long the model should be trained
    model_type
        A string with the type of model to use. Either 'resnet', 'efficient_net' or 'vgg19'.
    use_regu
        A boolean indicating whether L1L2 regularization should be used on the last layer.
    use_bias
        A boolean for adding a bias to the last layer.
    force_overfit
        A boolean for if the training schedule to be used should try to overfit the model or not.
    epochs_to_save
        A list of integers for the eventual saving of the model's checkpoints and training information.
        Useful for computing influence-related quantities using TracIn.
    verbose
        A boolean indicating whether to print the progress to stdout.
    use_tensorboard
        A boolean indicating the use of tensorboard for logging.
    """
    def __init__(
            self,
            epochs: int = 60,
            model_type: str = 'resnet',
            use_regu: bool = True,
            use_bias: bool = True,
            force_overfit: bool = False,
            epochs_to_save: Optional[List[int]] = None,
            verbose: bool = True,
            use_tensorboard: bool = False
    ):
        self.epochs = epochs
        self.model_type = model_type
        self.use_regu = use_regu
        self.use_bias = use_bias
        self.force_overfit = force_overfit
        self.epochs_to_save = epochs_to_save
        self.verbose = verbose
        self.use_tensorboard = use_tensorboard
        # first order
        self.schedule = [(10, 0.5), (20, 0.5), (30, 0.5), (40, 0.5), (50, 0.5)]
        self.learning_rate = 1E-3

    def train(
            self,
            training_dataset: tf.data.Dataset,
            test_dataset: tf.data.Dataset,
            train_batch_size: int = 128,
            test_batch_size: int = 128,
            log_path: Optional[str] = None
    ) -> Tuple[float, float, tf.keras.Model, Any]:
        """
        Trains the desired model with the object's training schedule on the provided training dataset, and
        validates its performance on the test dataset. Optionally, it logs the results progressively
        using the tensorboard framework.

        Parameters
        ----------
        training_dataset
            A TF dataset containing the samples on which the model will be trained. Typically, for this
            benchmark, it will be a noisy version of CIFAR-10.
        test_dataset
            A TF dataset with the test data on which to validate the model's accuracy. This data should
            only contain clean samples (i.e. no flipped labels).
        train_batch_size
            An integer specifying the batch size for the training procedure.
        test_batch_size
            An integer specifying the batch size for the validation procedure.
        log_path
            An (optional) string with the path onto which to save the tensorboard logs.

        Returns
        -------
        train_stats, test_stats, model, model_saver_information
            A tuple with the model's train accuracy, test accuracy, final model and the model saver information set to
            None for now
        """
        random_translation = tf.keras.layers.RandomTranslation(0.1, 0.1, fill_mode="nearest")

        def preprocess(x):
            x = tf.image.random_flip_left_right(x)
            x = random_translation(x)

            return x

        training_dataset = training_dataset.shuffle(1000).batch(train_batch_size)
        training_dataset_augment = training_dataset.map(lambda x, y: (preprocess(x), y))
        training_dataset_augment = training_dataset_augment.prefetch(100)

        test_dataset = test_dataset.batch(test_batch_size).prefetch(100)

        model = ConvNetCIFAR(self.model_type, self.use_regu, self.use_bias)

        loss = CategoricalCrossentropy(from_logits=True)

        if self.force_overfit:
            def LearningRateSchedulerMaxMin(lr_start=0.01, lr_end=0.0001, epochs=100):
                lr_decay = (lr_end / lr_start) ** (1. / epochs)
                return tf.keras.callbacks.LearningRateScheduler(lambda e: lr_start * lr_decay ** e)

            lr_scheduler = LearningRateSchedulerMaxMin(0.1, 0.0001, epochs=200)
            callbacks = [lr_scheduler]
            optimizer = tf.keras.optimizers.SGD(0.1, momentum=0.9)
        else:
            def lr_schedule(epoch: int):
                lr = self.learning_rate
                for ep_max, coeff in self.schedule:
                    if epoch >= ep_max:
                        lr *= coeff
                    else:
                        break
                return lr

            lr_scheduler = tf.keras.callbacks.LearningRateScheduler(lr_schedule)

            lr_reducer = tf.keras.callbacks.ReduceLROnPlateau(factor=np.sqrt(0.1),
                                                              cooldown=1,
                                                              patience=3,
                                                              min_lr=0.5e-6)
            callbacks = [lr_reducer, lr_scheduler]
            optimizer = Adam(learning_rate=1e-3)

        model.compile(optimizer=optimizer, loss=loss, metrics=['accuracy'])

        model_saver = None
        if self.epochs_to_save is not None:
            model_saver = ModelsSaver(self.epochs_to_save, optimizer)
            callbacks = [model_saver] + callbacks

        if self.use_tensorboard:
            if log_path is None:
                log_path = "./"
            callbacks.append(tf.keras.callbacks.TensorBoard(log_path))

        model.fit(training_dataset_augment, epochs=self.epochs, verbose=1 if self.verbose else 0,
                  validation_data=test_dataset,
                  callbacks=callbacks)
        _, train_stats = model.evaluate(training_dataset, batch_size=train_batch_size, verbose=0)
        _, test_stats = model.evaluate(test_dataset, batch_size=test_batch_size, verbose=0)

        if self.epochs_to_save is not None:
            return train_stats, test_stats, model, (model_saver.models, model_saver.learning_rates)

        return train_stats, test_stats, model, None


class Cifar10MislabelingDetectorEvaluator(MislabelingDetectorEvaluator):
    """
    An implementation of the mislabeled sample detection benchmark on the CIFAR10 image classification
    dataset. Useful for streamlining the evaluation of the different influence calculator techniques.

    Parameters
    ----------
    epochs
        An integer indicating for how long the model should be trained
    model_type
        A string with the type of model to use. Either 'resnet', 'efficient_net' or 'vgg19'.
    mislabeling_ratio
        A float with the ratio of noise to add to the training dataset's labels (ranging from 0. to 1.).
    use_regu
        A boolean indicating whether L1L2 regularization should be used on the last layer.
    force_overfit
        A boolean for if the training schedule to be used should try to overfit the model or not.
    train_batch_size
        An integer with the size of the batches on which to train the model.
    test_batch_size
        An integer with the size of the batches on which to perform the validation.
    influence_batch_size
        An integer with the size of the batches for performing the different operations
    epochs_to_save
        A list of integers for the eventual saving of the model's checkpoints and training information.
        Useful for computing influence-related quantities using TracIn.
    take_batch
        An optional integer indicating the number of batches to take if only one part of the
        training and test datasets are to be used.
    verbose_training
        A boolean indicating if progress about the training procedures should be reported to stdout.
    use_tensorboard
        A boolean indicating the use of tensorboard for logging.
    """
    def __init__(
            self,
            epochs: int = 60,
            model_type: str = 'resnet',
            mislabeling_ratio: float = 0.0005,
            use_regu: bool = True,
            use_bias: bool = True,
            force_overfit: bool = False,
            train_batch_size: int = 128,
            test_batch_size: int = 128,
            influence_batch_size: Optional[int] = None,
            epochs_to_save: Optional[List[int]] = None,
            take_batch: Optional[int] = None,
            verbose_training: bool = True,
            use_tensorboard: bool = False
        ): # pylint: disable=R0913

        config = {
            "epochs": epochs,
            "model_type": model_type,
            "mislabeling_ratio": mislabeling_ratio,
            "use_regularization": use_regu,
            "use_bias": use_bias,
            "optimizer": 'sgd' if force_overfit else 'adam',
            "train_batch_size": train_batch_size,
            "test_batch_size": test_batch_size,
            "epochs_to_save": epochs_to_save if epochs_to_save is not None else [],
            "samples_to_take": take_batch if take_batch is not None else -1,
        }

        (x_train, y_train), (x_test, y_test) = tf.keras.datasets.cifar10.load_data()
        x_train = tf.cast(2. * x_train / 255. - 1., dtype=tf.float32)
        x_test = tf.cast(2. * x_test / 255. - 1., dtype=tf.float32)

        y_train = tf.cast(tf.squeeze(tf.one_hot(y_train, depth=10, axis=1), axis=-1), dtype=tf.float32)
        y_test = tf.cast(tf.squeeze(tf.one_hot(y_test, depth=10, axis=1), axis=-1), dtype=tf.float32)

        training_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train))
        test_dataset = tf.data.Dataset.from_tensor_slices((x_test, y_test))

        if take_batch is not None:
            training_dataset = training_dataset.take(take_batch)
            test_dataset = test_dataset.take(take_batch)
        training_procedure = Cifar10TrainingProcedure(epochs, model_type, use_regu, use_bias, force_overfit,
                                                      epochs_to_save, verbose_training, use_tensorboard)
        super().__init__(training_dataset, test_dataset, training_procedure,
                         nb_classes=10, mislabeling_ratio=mislabeling_ratio,
                         train_batch_size=train_batch_size,
                         test_batch_size=test_batch_size,
                         influence_batch_size=influence_batch_size,
                         config=config)



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Utility classes and functions
"""
from .conjugate_gradients import conjugate_gradients_solve
from .backtracking_line_search import BacktrackingLineSearch
from .tf_operations import (
     find_layer,
     from_layer_name_to_layer_idx,
     is_dataset_batched,
     assert_batched_dataset,
     dataset_size,
     default_process_batch,
     dataset_to_tensor,
     array_to_dataset,
     map_to_device,
     split_model,
)
from .sorted_dict import BatchSort, ORDER
from .nearest_neighbors import BaseNearestNeighbors, LinearNearestNeighbors



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import setuptools
from setuptools import setup

with open("README.md", encoding="utf-8") as fh:
    README = fh.read()

setup(
    name="Influenciae",
    version="0.3.0",
    description="A Tensorflow Toolbox for Influence Functions",
    long_description=README,
    long_description_content_type="text/markdown",
    author="DEEL Core Team",
    author_email="agustin-martin.picard@irt-saintexupery.com",
    license="MIT",
    install_requires=['tensorflow >=2.7.0, <2.10.0', 'numpy', 'matplotlib'],
    extras_require={
        "tests": ["pytest", "pylint"],
        "docs": ["mkdocs", "mkdocs-material", "numkdoc"],
    },
    packages=setuptools.find_namespace_packages(include=["deel.*"]),
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Custom operations related to tensorflow objects
"""

import numpy as np
import tensorflow as tf

from ..types import Union, Tuple, Optional, Callable


def find_layer(model: tf.keras.Model, layer: Union[str, int]) -> tf.keras.layers.Layer:
    """
    Finds a layer in a model either by its name or by its index.

    Parameters
    ----------
    model
        Model on which to search.
    layer
        Layer name or layer index

    Returns
    -------
    layer
        Layer found
    """
    if isinstance(layer, str):
        return model.get_layer(layer)
    if isinstance(layer, int):
        return model.layers[layer]
    raise ValueError(f"Could not find any layer {layer}.")


def from_layer_name_to_layer_idx(model: tf.keras.Model, layer_name: str) -> int:
    """
    Finds the layer index of the corresponding layer name for the model

    Parameters
    ----------
    model
        Model on which to search
    layer_name
        Layer name

    Returns
    -------
    layer_idx
        The idx of the corresponding layer
    """
    for layer_idx, layer in enumerate(model.layers):
        if layer.name == layer_name:
            return layer_idx
    raise ValueError(f'No such layer: {layer_name}. Existing layers are: '
                     f'{list(layer.name for layer in model.layers)}.')

def split_model(model: tf.keras.Model, target_layer: Union[str, int]) -> Tuple[tf.keras.Model, tf.keras.Model]:
    """
    Splits a model into two sub-models, one containing the layers up to the target layer, and the other containing
    the layers from the target layer onwards.

    Parameters
    ----------
    model
        Model to split
    target_layer
        Layer name or layer index

    Returns
    -------
    model_1
        Model containing the layers up to the target layer
    model_2
        Model containing the layers from the target layer onwards
    """
    cloned_model = tf.keras.models.clone_model(model)
    cloned_model.set_weights(model.get_weights())
    cut_layer = find_layer(cloned_model, target_layer)
    model_1 = tf.keras.Model(inputs=cloned_model.inputs, outputs=cut_layer.input)
    model_2 = tf.keras.Model(inputs=tf.keras.Input(tensor=cut_layer.input), outputs=cloned_model.outputs)

    return model_1, model_2


def is_dataset_batched(dataset: tf.data.Dataset) -> Union[int, bool]:
    """
    Ensure the dataset is batched, if true return the batch_size else false.

    Parameters
    ----------
    dataset
        Tensorflow dataset to check.

    Returns
    -------
    batch_size
        False if the dataset if not batched else the batch_size of the dataset.
    """
    if hasattr(dataset, '_batch_size'):
        return dataset._batch_size  # pylint: disable=W0212

    return False


def assert_batched_dataset(dataset: tf.data.Dataset):
    """
    Throw an error if the dataset is not batched.

    Parameters
    ----------
    dataset
        Tensorflow dataset to check.
    """
    if not is_dataset_batched(dataset):
        raise ValueError("The dataset must be batched before performing this operation.")


def dataset_size(dataset: tf.data.Dataset):
    """
    Compute the size of a batched tensorflow dataset: batch_size * number of batchs

    Parameters
    ----------
    dataset
        Tensorflow dataset to check.

    Returns
    -------
    size
        Number of points in the dataset
    """
    assert_batched_dataset(dataset)

    size = dataset.cardinality().numpy() * dataset._batch_size  # pylint: disable=W0212
    return size


def default_process_batch(batch: Tuple[tf.Tensor, ...]) -> Tuple[tf.Tensor, tf.Tensor, Union[tf.Tensor, None]]:
    """
    A processing function to get the information into the right format for ingestion in the case of
    the default case where the batch is in the format (x, y) and there's no weight in a per-sample basis.

    Parameters
    ----------
    batch
        A tuple of tensors, where the information is in the format (x, y).

    Returns
    -------
    processed_batch
        A tuple containing explicitly the inputs, the targets and the per-sample weights (if present, None otherwise)
    """
    y_true = batch[1]
    model_inp = batch[0]
    sample_weight = None

    return model_inp, y_true, sample_weight


def dataset_to_tensor(dataset: tf.data.Dataset) -> tf.Tensor:
    """
    Transforms a (batched) dataset into a tensor for easier manipulation.

    Parameters
    ----------
    dataset
        A batched TF dataset to transform into a tensor.

    Returns
    -------
    tensor
        A TF tensor with the evaluated dataset.
    """
    assert_batched_dataset(dataset)
    if isinstance(dataset.element_spec, Tuple):
        tensor = [list(dataset.map(lambda *w: w[i]).unbatch()) for i in range(len(dataset.element_spec))]  # pylint: disable=W0640
    else:
        tensor = tf.concat(list(dataset), axis=0)

    return tensor


def extract_only_values(dataset: tf.data.Dataset) -> tf.Tensor:
    """
    Utility function to extract all the influence values from a dataset containing tuples (data-point, influence score).
    The dataset much be batched before applying this operation.

    Parameters
    ----------
    dataset
        A batched TF dataset with tuples (data-point, influence score).

    Returns
    -------
    influence_values
        A tensor with only the dataset's influence values.
    """
    return dataset_to_tensor(dataset.map(lambda *inputs: inputs[1]))


def array_to_dataset(
        array: Union[np.ndarray, tf.Tensor, Tuple[np.ndarray, ...], Tuple[tf.Tensor, ...]],
        batch_size: Optional[int] = None,
        shuffle: bool = False,
        buffer_size: int = 1000
) -> tf.data.Dataset:
    """
    Converts a dataset in the form of a numpy array or tensor (in tuples when it contains the labels) into a
    TF dataset, and if desired, shuffles and batches it for easier ingestion by other functions.

    Parameters
    ----------
    array
        Either a numpy array (or tensor) with the samples or a tuple with a set of samples and labels (or value
        associated to them).
    batch_size
        An optional integer indicating the batch size. If None, the dataset won't be batched
    shuffle
        A boolean indicating whether the dataset should be shuffled.
    buffer_size
        An integer for both the shuffle operation and computing the prefetch size (when batching).

    Returns
    -------
    dataset
        A TF dataset with the data in the array, possibly shuffled, batched and prefetched.
    """
    if isinstance(array, np.ndarray):
        array = tf.convert_to_tensor(array, dtype=tf.float32)
    elif isinstance(array, Tuple) and isinstance(array[0], np.ndarray):
        array = tf.stack(list(
            zip(tf.convert_to_tensor(array[0], dtype=tf.float32), tf.convert_to_tensor(array[1], dtype=tf.float32))
        ), axis=0)

    dataset = tf.data.Dataset.from_tensor_slices(array)
    if shuffle:
        dataset = dataset.shuffle(buffer_size)
    if batch_size is not None:
        assert batch_size > 0
        assert buffer_size > 0
        dataset = dataset.batch(batch_size).prefetch(buffer_size // batch_size)

    return dataset


def get_device(device: Optional[str]) -> str:
    """
    Gets the name of the device to use. If there are any available GPUs, it will use the first one
    in the system, otherwise, it will use the CPU.

    Parameters
    ----------
    device
        A string specifying the device on which to run the computations. If None, it will search
        for available GPUs, and if none are found, it will return the first CPU.

    Returns
    -------
    device
        A string with the name of the device on which to run the computations.
    """
    if device is not None:
        return device

    physical_devices = tf.config.list_physical_devices('GPU')
    if physical_devices is None or len(physical_devices) == 0:
        return 'cpu:0'
    return 'GPU:0'


def map_to_device(dataset: tf.data.Dataset, map_fun: Callable, device: Optional[str] = None) -> tf.data.Dataset:
    """
    Performs a map function on the preferred device. If none is specified, the first available GPU is
    chosen, and if there aren't any, the computations are done on the CPU.

    Parameters
    ----------
    dataset
        The dataset on which to apply the map function.
    map_fun
        A callable with the function to be applied to the whole CPU.
    device
        An (optional) string with the device on which to perform the map function.
        If none is specified, the first available GPU is chosen, and if there aren't any, the
        computations are done on the CPU.

    Returns
    -------
    mapped_dataset
        The dataset with the map function applied to it.
    """
    device = get_device(device)

    def map_fun_device(*args):
        with tf.device(device):
            result = map_fun(*args)
        return result

    return dataset.map(map_fun_device)



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Module implementing a data-point-influence estimation method based on how intensely
the model's weights must be adversarially attacked to make it change its prediction
for each sample. Intuitively, outliers and atypical examples will need a lighter
deformation of the boundary for it to place them in the wrong class.

Unlike other influence calculators, this one cannot be used to estimate
the influence of a point on another.

This boundary deformation process is performed using deep fool on the target weights.
"""
import tensorflow as tf
from tensorflow.keras import Model  # pylint:  disable=E0611

from ..common import SelfInfluenceCalculator
from ..types import Tuple, List


class WeightsBoundaryCalculator(SelfInfluenceCalculator):
    """
    A class implementing an influence score based on the distance of a sample to the boundary of its classifier.
    The distance to the boundary is estimated by deforming the boundary of the model to move a given sample
    to the closest adversarial class.
    To compute this distance, the deep fool method is used on the weights of the model (deep fool originally compute
    the distance on the sample space).
    [https://arxiv.org/abs/1511.04599]

    Notes
    -----
    This method has better mislabeled-sample-detection performance when the model overfits.

    Parameters
    ----------
    model
        A TF2 model that has already been trained
    step_nbr
        Number of the iterations to find the closest adversarial problem
    norm_type
        The distance norm used to compute the distance to the boundary
    eps
        Difference between two logits to assume that the logits have the same values
    """

    def __init__(self, model: Model, step_nbr: int = 100, norm_type: int = 2, eps: float = 1E-6):
        self.weights_init = [tf.identity(w) for w in model.trainable_variables]
        self.model = model

        self.step_nbr = step_nbr
        self.norm_type = norm_type
        self.eps = eps

    def __compute_norm(self, weights: List[tf.Tensor]) -> tf.float32:
        """
        Compute the norm of a list of weights

        Parameters
        ----------
        weights
            The list of weights

        Returns
        -------
        weights_norm
            The norm of the weights
        """
        weights_flatten = tf.concat([tf.reshape(w, (-1,)) for w in weights], axis=0)
        weights_norm = tf.norm(weights_flatten, ord=self.norm_type)

        return weights_norm

    def __delta_to_index(self, indexes_1: tf.Tensor, indexes_2: tf.Tensor, x: tf.Tensor) -> tf.Tensor:
        """
        Compute the difference between the logit of a given class and the other logits

        Parameters
        ----------
        indexes_1
            The logits of other classes
        indexes_2
            The logits of the predicted class
        x
            The logits

        Returns
        -------
        delta_x
            The difference between the logits
        """
        x1 = tf.gather(x, indexes_1, batch_dims=1)
        x2 = tf.gather(x, tf.expand_dims(indexes_2, axis=1), batch_dims=1)

        delta_x = x1 - tf.repeat(x2, tf.shape(x1)[1], axis=1)

        return delta_x

    @tf.function
    def _step(self, x: tf.Tensor, y_pred: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        The optimization step to find the distance between the boundary and a given sample x.
        To see more details about the optimization procedure for multi-class classifiers,
        please refer to [https://arxiv.org/abs/1511.04599]

        Notes
        -----
        This function updates the weights of the model at each step.

        Parameters
        ----------
        x
            The current sample used to compute the distance to the boundary of the model
        y_pred
            The one-hot labels predicted by the current model for the x sample

        Returns
        -------
        computation
            Boolean to determine if the optimization process should continue.
            True if the sample didn't change of class, False if not.
        loss_value
            The loss of the optimization procedure
        """
        y_pred = tf.argmax(y_pred, axis=1)

        with tf.GradientTape() as tape:
            y = self.model(x)

        y_computed = tf.argmax(y, axis=1)

        def update_grads():
            model_weights = self.model.trainable_variables
            jac = tape.jacobian(y, self.model.trainable_variables)

            indexes_all = tf.repeat(tf.expand_dims(tf.range(0, tf.shape(y)[1]), axis=0), tf.shape(y)[0], axis=0)
            indexes_class = tf.cast(tf.repeat(tf.expand_dims(y_pred, axis=1), tf.shape(y)[1], axis=1), dtype=tf.int32)
            indexes_other = tf.reshape(indexes_all[indexes_all != indexes_class], (-1, tf.shape(y)[1] - 1))

            delta_y = self.__delta_to_index(indexes_other, y_pred, y)
            delta_y = tf.abs(tf.reduce_mean(delta_y, axis=0))

            jac_delta = [tf.reduce_mean(self.__delta_to_index(indexes_other, y_pred, j), axis=0) for j in jac]

            jac_norm = tf.concat([tf.reshape(j, (tf.shape(j)[0], -1)) for j in jac_delta], axis=1)
            jac_norm = tf.norm(jac_norm, axis=1, ord=self.norm_type)

            coeff = delta_y / jac_norm

            best_class = tf.argmin(coeff, axis=0)

            loss = (coeff / tf.pow(jac_norm, self.norm_type - 1))[best_class]

            weights = [w + loss * tf.pow(tf.abs(g[best_class]), self.norm_type - 1) * tf.sign(g[best_class]) for w, g in
                       zip(model_weights, jac_delta)]

            for m, w in zip(model_weights, weights):
                m.assign(w)

            return loss

        computation = tf.reduce_any(y_computed == y_pred)

        top_k, _ = tf.math.top_k(tf.squeeze(y, axis=0), k=2)
        is_close_enough = tf.abs(top_k[0] - top_k[1]) > self.eps

        computation = tf.logical_and(computation, is_close_enough)

        loss_value = tf.cond(computation, update_grads, lambda: tf.constant(0.0))

        return computation, loss_value

    def __delta_weights(self) -> tf.float32:
        """
        Compute the norm between the trained weights of the model and the current modified weights of the model.

        Returns
        -------
        norm
            The distance between the initial model and the current model
        """
        weights = [w1 - w2 for w1, w2 in zip(self.weights_init, self.model.trainable_variables)]
        norm = self.__compute_norm(weights)

        return norm

    def __reset_weights(self) -> None:
        """
        Set the weights of the model to the initial trained weights.
        """
        for m, w in zip(self.model.trainable_variables, self.weights_init):
            m.assign(w)

    def __compute_single_sample_score(self, x: tf.Tensor) -> tf.Tensor:
        """
        Computes the influence score (self-influence) for a single training samples.

        Parameters
        ----------
        x
            A tensor with a single training sample.

        Returns
        -------
        score
            The influence score of the sample.
        """

        x = tf.expand_dims(x, axis=0)
        y_pred = self.model(x)

        tf.while_loop(lambda cond, index: tf.logical_and(cond, index < self.step_nbr),
                      lambda cond, index: (self._step(x, y_pred)[0], index + 1),
                      [tf.constant(True), tf.constant(0, dtype=tf.int32)])

        score = self.__delta_weights()

        self.__reset_weights()

        return score

    def _compute_influence_value_from_batch(self, train_samples: Tuple[tf.Tensor, ...]) -> tf.Tensor:
        """
        Computes the influence score (self-influence) for a single batch of training samples.

        Parameters
        ----------
        train_samples
            A tensor with a single batch of training sample.

        Returns
        -------
        influence_values
            The influence score of each sample in the batch train_samples.
        """
        scores = tf.map_fn(self.__compute_single_sample_score, train_samples[:-1][0], parallel_iterations=1)
        scores = - tf.expand_dims(scores, axis=1)

        return scores



# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""ResNet v1, v2, and segmentation models for Keras.

# Reference

- [Deep Residual Learning for Image Recognition](https://arxiv.org/abs/1512.03385)
- [Identity Mappings in Deep Residual Networks](https://arxiv.org/abs/1603.05027)

Reference material for extended functionality:

- [ResNeXt](https://arxiv.org/abs/1611.05431) for Tiny ImageNet support.
- [Dilated Residual Networks](https://arxiv.org/pdf/1705.09914) for segmentation support
- [Deep Residual Learning for Instrument Segmentation in
   Robotic Surgery](https://arxiv.org/abs/1703.08580)
  for segmentation support.

Implementation Adapted from: github.com/raghakot/keras-resnet
Implementation Adapted from: github.com/keras-team/keras-contrib
"""
import six

from tensorflow.keras import Model  #pylint:  disable=E0611
from tensorflow.keras import backend as K  #pylint:  disable=E0611
from tensorflow.keras.regularizers import l2  #pylint:  disable=E0611
from tensorflow.keras.layers import ( #pylint:  disable=E0611
    Input,
    Activation,
    Reshape,
    ReLU,
    Dense,
    Conv2D,
    MaxPooling2D,
    GlobalMaxPooling2D,
    GlobalAveragePooling2D,
    Dropout,
    Add,
    BatchNormalization
)


def _bn_relu(x, bn_name=None, relu_name=None):
    """
    Helper to build a BN -> relu block
    """
    if K.image_data_format() == 'channels_last':
        CHANNEL_AXIS = 3
    else:
        CHANNEL_AXIS = 1
    norm = BatchNormalization(axis=CHANNEL_AXIS, name=bn_name)(x)
    return Activation("relu", name=relu_name)(norm)


def get_Conv2D(filters, kernel_size=(3, 3), padding='circular', activation=ReLU, use_bias=True, strides=(1, 1),
               name=None):
    """
    Helper to build Convolutional layers
    """
    return Conv2D(filters=filters, kernel_size=kernel_size, padding=padding, activation=activation,
                   use_bias=use_bias, strides=strides, name=name)


def _conv_bn_relu(**conv_params):
    """
    Helper to build a conv -> BN -> relu residual unit activation function.
    This is the original ResNet v1 scheme in https://arxiv.org/abs/1512.03385
    """
    filters = conv_params["filters"]
    kernel_size = conv_params["kernel_size"]
    strides = conv_params.setdefault("strides", (1, 1))
    dilation_rate = conv_params.setdefault("dilation_rate", (1, 1))
    conv_name = conv_params.setdefault("conv_name", None)
    bn_name = conv_params.setdefault("bn_name", None)
    relu_name = conv_params.setdefault("relu_name", None)
    kernel_initializer = conv_params.setdefault("kernel_initializer", "he_normal")
    padding = conv_params.setdefault("padding", "same")
    kernel_regularizer = conv_params.setdefault("kernel_regularizer", l2(1.e-4))

    def f(x):
        x = Conv2D(filters=filters, kernel_size=kernel_size,
                       strides=strides, padding=padding,
                       dilation_rate=dilation_rate,
                       kernel_initializer=kernel_initializer,
                       kernel_regularizer=kernel_regularizer,
                       name=conv_name)(x)

        return _bn_relu(x, bn_name=bn_name, relu_name=relu_name)

    return f


def _bn_relu_conv(**conv_params):
    """
    Helper to build a BN -> relu -> conv residual unit with full pre-activation
    function. This is the ResNet v2 scheme proposed in
    http://arxiv.org/pdf/1603.05027v2.pdf
    """
    filters = conv_params["filters"]
    kernel_size = conv_params["kernel_size"]
    strides = conv_params.setdefault("strides", (1, 1))
    dilation_rate = conv_params.setdefault("dilation_rate", (1, 1))
    conv_name = conv_params.setdefault("conv_name", None)
    bn_name = conv_params.setdefault("bn_name", None)
    relu_name = conv_params.setdefault("relu_name", None)
    kernel_initializer = conv_params.setdefault("kernel_initializer", "he_normal")
    padding = conv_params.setdefault("padding", "same")
    kernel_regularizer = conv_params.setdefault("kernel_regularizer", l2(1.e-4))

    def f(x):
        activation = _bn_relu(x, bn_name=bn_name, relu_name=relu_name)
        return Conv2D(filters=filters, kernel_size=kernel_size,
                          strides=strides, padding=padding,
                          dilation_rate=dilation_rate,
                          kernel_initializer=kernel_initializer,
                          kernel_regularizer=kernel_regularizer,
                          name=conv_name)(activation)

    return f


def _shortcut(input_feature, residual, conv_name_base=None, bn_name_base=None):
    """
    Adds a shortcut between input and residual block and merges them with "sum"
    """
    # Expand channels of shortcut to match residual.
    # Stride appropriately to match residual (width, height)
    # Should be int if network architecture is correctly configured.
    input_shape = K.int_shape(input_feature)
    residual_shape = K.int_shape(residual)
    # Figure out the ordering of the channels
    if K.image_data_format() == 'channels_last':
        ROW_AXIS = 1
        COL_AXIS = 2
        CHANNEL_AXIS = 3
    else:
        CHANNEL_AXIS = 1
        ROW_AXIS = 2
        COL_AXIS = 3
    stride_width = int(round(input_shape[ROW_AXIS] / residual_shape[ROW_AXIS]))
    stride_height = int(round(input_shape[COL_AXIS] / residual_shape[COL_AXIS]))
    equal_channels = input_shape[CHANNEL_AXIS] == residual_shape[CHANNEL_AXIS]

    shortcut = input_feature
    # 1 X 1 conv if shape is different. Else identity.
    if stride_width > 1 or stride_height > 1 or not equal_channels:

        if conv_name_base is not None:
            conv_name_base = conv_name_base + '1'
        shortcut = Conv2D(filters=residual_shape[CHANNEL_AXIS],
                          kernel_size=(1, 1),
                          strides=(stride_width, stride_height),
                          padding="valid",
                          kernel_initializer="he_normal",
                          kernel_regularizer=l2(0.0001),
                          name=conv_name_base)(input_feature)
        if bn_name_base is not None:
            bn_name_base = bn_name_base + '1'
        shortcut = BatchNormalization(axis=CHANNEL_AXIS,
                                      name=bn_name_base)(shortcut)

    return Add()([shortcut, residual])


def _residual_block(block_function, filters, blocks, stage,
                    transition_strides=None, transition_dilation_rates=None,
                    dilation_rates=None, is_first_layer=False, dropout=None,
                    residual_unit=_bn_relu_conv):
    """
    Builds a residual block with repeating bottleneck blocks.

    stage: integer, current stage label, used for generating layer names
    blocks: number of blocks 'a','b'..., current block label, used for generating
            layer names
    transition_strides: a list of tuples for the strides of each transition
    transition_dilation_rates: a list of tuples for the dilation rate of each
            transition
    """
    if transition_dilation_rates is None:
        transition_dilation_rates = [(1, 1)] * blocks
    if transition_strides is None:
        transition_strides = [(1, 1)] * blocks
    if dilation_rates is None:
        dilation_rates = [1] * blocks

    def f(x):
        for i in range(blocks):
            is_first_block = is_first_layer and i == 0
            x = block_function(filters=filters, stage=stage, block=i,
                               transition_strides=transition_strides[i],
                               dilation_rate=dilation_rates[i],
                               is_first_block_of_first_layer=is_first_block,
                               dropout=dropout,
                               residual_unit=residual_unit)(x)
        return x

    return f


def _block_name_base(stage, block):
    """
    Get the convolution name base and batch normalization name base defined by
    stage and block.

    If there are less than 26 blocks they will be labeled 'a', 'b', 'c' to match the
    paper and keras and beyond 26 blocks they will simply be numbered.
    """
    if block < 27:
        block = '%c' % (block + 97)  # 97 is the ascii number for lowercase 'a'  # pylint: disable=C0209
    conv_name_base = f'res{stage}{block}_branch'
    bn_name_base = f'bn{stage}{block}_branch'
    return conv_name_base, bn_name_base


def basic_block(filters, stage, block, transition_strides=(1, 1),
                dilation_rate=(1, 1), is_first_block_of_first_layer=False, dropout=None,
                residual_unit=_bn_relu_conv):
    """
    Basic 3 X 3 convolution blocks for use on resnets with layers <= 34.
    Follows improved proposed scheme in http://arxiv.org/pdf/1603.05027v2.pdf
    """

    def f(input_features):
        conv_name_base, bn_name_base = _block_name_base(stage, block)
        if is_first_block_of_first_layer:
            # don't repeat bn->relu since we just did bn->relu->maxpool
            x = Conv2D(filters=filters, kernel_size=(3, 3),
                       strides=transition_strides,
                       dilation_rate=dilation_rate,
                       padding="same",
                       kernel_initializer="he_normal",
                       kernel_regularizer=l2(1e-4),
                       name=conv_name_base + '2a')(input_features)
        else:
            x = residual_unit(filters=filters, kernel_size=(3, 3),
                              strides=transition_strides,
                              dilation_rate=dilation_rate,
                              conv_name_base=conv_name_base + '2a',
                              bn_name_base=bn_name_base + '2a')(input_features)

        if dropout is not None:
            x = Dropout(dropout)(x)

        x = residual_unit(filters=filters, kernel_size=(3, 3),
                          conv_name_base=conv_name_base + '2b',
                          bn_name_base=bn_name_base + '2b')(x)

        return _shortcut(input_features, x)

    return f


def bottleneck(filters, stage, block, transition_strides=(1, 1),
               dilation_rate=(1, 1), is_first_block_of_first_layer=False, dropout=None,
               residual_unit=_bn_relu_conv):
    """
    Bottleneck architecture for > 34 layer resnet.
    Follows improved proposed scheme in http://arxiv.org/pdf/1603.05027v2.pdf

    Returns
    -------
        A final conv layer of filters * 4
    """

    def f(input_feature):
        conv_name_base, bn_name_base = _block_name_base(stage, block)
        if is_first_block_of_first_layer:
            # don't repeat bn->relu since we just did bn->relu->maxpool
            x = Conv2D(filters=filters, kernel_size=(1, 1),
                       strides=transition_strides,
                       dilation_rate=dilation_rate,
                       padding="same",
                       kernel_initializer="he_normal",
                       kernel_regularizer=l2(1e-4),
                       name=conv_name_base + '2a')(input_feature)
        else:
            x = residual_unit(filters=filters, kernel_size=(1, 1),
                              strides=transition_strides,
                              dilation_rate=dilation_rate,
                              conv_name_base=conv_name_base + '2a',
                              bn_name_base=bn_name_base + '2a')(input_feature)

        if dropout is not None:
            x = Dropout(dropout)(x)

        x = residual_unit(filters=filters, kernel_size=(3, 3),
                          conv_name_base=conv_name_base + '2b',
                          bn_name_base=bn_name_base + '2b')(x)

        if dropout is not None:
            x = Dropout(dropout)(x)

        x = residual_unit(filters=filters * 4, kernel_size=(1, 1),
                          conv_name_base=conv_name_base + '2c',
                          bn_name_base=bn_name_base + '2c')(x)

        return _shortcut(input_feature, x)

    return f


def _string_to_function(identifier):
    if isinstance(identifier, six.string_types):
        res = globals().get(identifier)
        if not res:
            raise ValueError(f'Invalid {identifier}')
        return res
    return identifier


def ResNet(input_shape=None, classes=10, block='bottleneck', residual_unit='v2',  # pylint: disable=R0912 disable=R0913 disable=R0915
           repetitions=None, initial_filters=64, activation='softmax', include_top=True,
           input_tensor=None, dropout=None, transition_dilation_rate=(1, 1),
           initial_strides=(1, 1), initial_kernel_size=(3, 3), initial_pooling='max',
           final_pooling='avg', top='classification'):
    """
    Builds a custom ResNet like architecture. Defaults to ResNet50 v2.

    Parameters
    ----------
        input_shape: optional shape tuple, only to be specified
            if `include_top` is False (otherwise the input shape
            has to be `(224, 224, 3)` (with `channels_last` dim ordering)
            or `(3, 224, 224)` (with `channels_first` dim ordering).
            It should have exactly 3 dimensions,
            and width and height should be no smaller than 8.
            E.g. `(224, 224, 3)` would be one valid value.
        classes: The number of outputs at final softmax layer
        block: The block function to use. This is either `'basic'` or `'bottleneck'`.
            The original paper used `basic` for layers < 50.
        repetitions: Number of repetitions of various block units.
            At each block unit, the number of filters are doubled and the input size
            is halved. Default of None implies the ResNet50v2 values of [3, 4, 6, 3].
        residual_unit: the basic residual unit, 'v1' for conv bn relu, 'v2' for bn relu
            conv. See [Identity Mappings in
            Deep Residual Networks](https://arxiv.org/abs/1603.05027)
            for details.
        dropout: None for no dropout, otherwise rate of dropout from 0 to 1.
            Based on [Wide Residual Networks.(https://arxiv.org/pdf/1605.07146) paper.
        transition_dilation_rate: Dilation rate for transition layers. For semantic
            segmentation of images use a dilation rate of (2, 2).
        initial_strides: Stride of the very first residual unit and MaxPooling2D call,
            with default (2, 2), set to (1, 1) for small images like cifar.
        initial_kernel_size: kernel size of the very first convolution, (7, 7) for
            imagenet and (3, 3) for small image datasets like tiny imagenet and cifar.
            See [ResNeXt](https://arxiv.org/abs/1611.05431) paper for details.
        initial_pooling: Determine if there will be an initial pooling layer,
            'max' for imagenet and None for small image datasets.
            See [ResNeXt](https://arxiv.org/abs/1611.05431) paper for details.
        final_pooling: Optional pooling mode for feature extraction at the final
            model layer when `include_top` is `False`.
            - `None` means that the output of the model
                will be the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a
                2D tensor.
            - `max` means that global max pooling will
                be applied.
        top: Defines final layers to evaluate based on a specific problem type. Options
            are 'classification' for ImageNet style problems, 'segmentation' for
            problems like the Pascal VOC dataset, and None to exclude these layers
            entirely.

    Returns
    -------
        The keras `Model`.
    """
    if activation not in ['softmax', 'sigmoid', None]:
        raise ValueError('activation must be one of "softmax", "sigmoid", or None')
    if activation == 'sigmoid' and classes != 1:
        raise ValueError('sigmoid activation can only be used when classes = 1')
    if repetitions is None:
        repetitions = [3, 4, 6, 3]
    # Determine proper input shape
    if len(input_shape) != 3:
        raise TypeError("Input shape should be a tuple (nb_channels, nb_rows, nb_cols)")

    if block == 'basic':
        block_fn = basic_block
    elif block == 'bottleneck':
        block_fn = bottleneck
    elif isinstance(block, six.string_types):
        block_fn = _string_to_function(block)
    else:
        block_fn = block

    if residual_unit == 'v2':
        residual_unit = _bn_relu_conv
    elif residual_unit == 'v1':
        residual_unit = _conv_bn_relu
    elif isinstance(residual_unit, six.string_types):
        residual_unit = _string_to_function(residual_unit)

    # Permute dimension order if necessary
    if K.image_data_format() == 'channels_first':
        input_shape = (input_shape[1], input_shape[2], input_shape[0])

    # Determine proper input shape
    img_input = Input(shape=input_shape, tensor=input_tensor)
    x = _conv_bn_relu(filters=initial_filters, kernel_size=initial_kernel_size,
                      strides=initial_strides)(
        img_input)
    if initial_pooling == 'max':
        x = MaxPooling2D(pool_size=(3, 3), strides=initial_strides, padding="same")(x)

    block = x
    filters = initial_filters
    for i, r in enumerate(repetitions):
        transition_dilation_rates = [transition_dilation_rate] * r
        transition_strides = [(1, 1)] * r
        if transition_dilation_rate == (1, 1) and i > 0:  ##FM do not stride on first layer
            transition_strides[0] = (2, 2)
        block = _residual_block(block_fn, filters=filters,
                                stage=i, blocks=r,
                                is_first_layer=(i == 0) and (residual_unit == 'v2'),
                                dropout=dropout,
                                transition_dilation_rates=transition_dilation_rates,
                                transition_strides=transition_strides,
                                residual_unit=residual_unit)(block)
        filters *= 2

    # Last activation
    x = _bn_relu(block)

    # Classifier block
    if include_top and top == 'classification':
        x = GlobalAveragePooling2D()(x)
        x = Dense(units=classes, activation=activation,
                  kernel_initializer="he_normal")(x)
    elif include_top and top == 'segmentation':
        x = Conv2D(classes, (1, 1), activation='linear', padding='same')(x)

        if K.image_data_format() == 'channels_first':
            _, row, col = input_shape
        else:
            row, col, _ = input_shape

        x = Reshape((row * col, classes))(x)
        x = Activation(activation)(x)
        x = Reshape((row, col, classes))(x)
    elif final_pooling == 'avg':
        x = GlobalAveragePooling2D()(x)
    elif final_pooling == 'max':
        x = GlobalMaxPooling2D()(x)

    model = Model(inputs=img_input, outputs=x)
    return model
